#
# Copyright (c) 2007 Patrick Van Oosterwijck
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#

"""Objects that are used in the game"""

import random
import pygame
import gameobj
import world


class Balloon(gameobj.GameObj):
    """Balloon class that takes care of defining the behavior of the
    player's balloon"""

    def __init__(self, *groups):
        """Init function when creating the object"""
        # Invoke base class init to load all object data
        gameobj.GameObj.__init__(self, 'balloon', 0.5, world.scr_scale, *groups)
        # Reset to default state
        self.Reset()
        # Set some physical limits
        self.max_vspeed_up = 2.0
        self.max_vspeed_down = 15.0
        self.max_hspeed = 1.0
        self.max_lift = 10.2
        self.gravity = 10.0

    def Reset(self):
        """Reset the balloon to its default state"""
        # Set the default state
        self.SetState('default')
        # Reset balloon object variables
        self.fuel_level = 100.0
        self.lift = 10.001
        self.position = 0.0, 0.0
        self.rect = self.image['default.png']['surface'].get_rect()
        self.collide_rect = self.rect.inflate(-self.rect.height*0.135,
                                              -self.rect.height*0.135)
        self.collide_rect.height *= 0.3
        self.save_collide_rect = self.collide_rect
        self.speed = 0.0, 0.0
        self.flame_on = False
        self.vent_open = False
        self.damaged = False
        self.static = False

    def SetFlameOn(self, on):
        """Turn the balloon flame on or off, controlled by user input"""
        if on and self.fuel_level > 0.0 and not self.damaged:
            self.flame_on = True
            self.SetState('flame')
        else:
            self.flame_on = False
            self.SetState('default')

    def SetVentOpen(self, setopen):
        """Open the vent or close it, controlled by user input"""
        self.vent_open = setopen

    def SetDamaged(self):
        """Set the balloon to damaged"""
        self.damaged = True
        self.SetState('damaged')

    def SetStatic(self):
        """Set the balloon static when not playing"""
        self.static = True

    def update(self):
        """Alter balloon speed and position based on external and
        internal factors"""
        # If in static mode, don't update balloon
        if self.static: return
        # Is the balloon damaged?
        if self.damaged:
            # Leak away our lift
            self.lift *= 0.9
        else:
            # Is the flame on?
            if self.flame_on:
                # Increase lift
                self.lift += min(0.0005, (self.max_lift - self.lift) / 50.0)
                # Decrease fuel
                self.fuel_level -= 0.5
                # Did we run out of fuel?
                if self.fuel_level <= 0.0:
                    # Limit to zero
                    self.fuel_level = 0.0
                    # Turn the flame off
                    self.SetFlameOn(False)
        # Apply lift decay
        self.lift *= 0.999999
        # Apply extra lift decay if the vent is open
        if self.vent_open: self.lift *= 0.9999
        # Apply wind speeds
        windspeed = 0.0
        for windlayer in pygame.sprite.spritecollide(
                    self, world.trench.wind_group, False):
            windspeed += windlayer.misc
        # Change speed based on lift-gravity vertically and wind speed
        # horizontally, and limit to maximum speed
        self.speed = (max(min(self.speed[0]*0.99+windspeed*0.001,
                              self.max_hspeed), -self.max_hspeed),
                      max(min(self.speed[1]*0.99+self.gravity-self.lift,
                              self.max_vspeed_down), -self.max_vspeed_up))
        # Change position based on speed
        self.position = (self.position[0]+self.speed[0],
                         self.position[1]+self.speed[1])
        self.rect.topleft = self.position
        self.collide_rect = self.save_collide_rect.move(self.position)


class Crow(gameobj.GameObj):
    """Crow class that takes care of defining the behavior of a crow"""

    def __init__(self, *groups):
        """Init function when creating the object"""
        # Invoke base class init to load all object data
        gameobj.GameObj.__init__(self, 'crow', 0.4, world.scr_scale, *groups)
        # Save the speed and lift factor
        self.speed = random.uniform(-1.5, 1.5)
        self.lift_factor = random.uniform(-0.1, 0.1)
        # Get the bounding rectangle
        self.rect = self.image['1.png']['surface'].get_rect()
        # Calculate collision rectangle
        self.collide_rect = self.rect.inflate(0.0, -self.rect.height*0.7)
        self.save_collide_rect = self.collide_rect
        # Depending on the direction, decide on which side of the screen
        # to start and which state to use, plus enforce minimum speed
        if self.speed > 0.0:
            self.speed += 0.5
            self.position = -self.rect.width, \
                    random.uniform(50, world.trench_depth-200)
            self.SetState('flyright')
        else:
            self.speed -= 0.5
            self.position = world.trench_width, \
                    random.uniform(50, world.trench_depth-200)
            self.SetState('flyleft')

    def update(self):
        """Calculate next crow position"""
        # Calculate new position
        self.position = (self.position[0]+self.speed,
                         self.position[1]-self.speed*self.lift_factor)
        self.rect.topleft = self.position
        self.collide_rect = self.save_collide_rect.move(self.position)
        # Check if we've traveled to the other side, get rid of the crow if so
        if self.speed > 0.0:
            if self.position[0] > world.trench_width:
                self.kill()
        else:
            if self.position[0] < -self.rect.width:
                self.kill()


class Parachute(gameobj.GameObj):
    """Parachute class defines the behavior of a parachute"""
    
    def __init__(self, *groups):
        """Init function when creating the object"""
        # Invoke base class init to load all object data
        gameobj.GameObj.__init__(self, 'parachute', 0.45,
                                 world.scr_scale, *groups)
        # Set the initial position
        self.position = random.uniform(
                    world.trench_width*0.1, world.trench_width*0.9), \
                    random.uniform(-world.view_height/world.scr_scale,
                    world.view_y-0.2*world.view_height/world.scr_scale)
        # Get the bounding rectangle
        self.rect = self.image['parachute.png']['surface'].get_rect()
        # Calculate collision rectangle
        self.collide_rect = self.rect.inflate(-self.rect.height*0.13,
                                              -self.rect.height*0.13)
        self.save_collide_rect = self.collide_rect
        # Set the default state
        self.SetState('default')
        # Set the initial speed
        self.speed = random.uniform(-0.1, 0.1), random.uniform(0.2, 1.2)
        # Set limit for horizontal speed
        self.max_hspeed = 1.0

    def update(self):
        """Calculate next parachute position"""
        # Apply wind speeds
        windspeed = 0.0
        for windlayer in pygame.sprite.spritecollide(
                    self, world.trench.wind_group, False):
            windspeed += windlayer.misc
        # Change horizontal speed randomly and based on wind speed,
        # and limit to maximum speed
        self.speed = (max(min(self.speed[0]*0.99 + windspeed*0.0001
                              + random.uniform(-0.0001, 0.0001),
                              self.max_hspeed), -self.max_hspeed),
                      self.speed[1])
        # Change position based on speed
        self.position = (self.position[0]+self.speed[0],
                         self.position[1]+self.speed[1])
        self.rect.topleft = self.position
        self.collide_rect = self.save_collide_rect.move(self.position)
        # Check if the parachute landed in the water, get rid of it if so
        if self.position[1] > world.trench_depth:
            self.kill()


class Flag(gameobj.GameObj):
    """Flag class defines behavior of a flag that indicates wind speed"""

    def __init__(self, wind, *groups):
        """Init function when creating the object"""
        # Invoke base class init to load all object data
        gameobj.GameObj.__init__(self, 'flag', 0.5,
                                 world.scr_scale, *groups)
        # Get the bounding rectangle
        self.rect = self.image['0.png']['surface'].get_rect()
        # Set the flag state based on wind speed and direction
        dircode = ['lr', 'rl']
        self.SetState('wind%d%s' % (int(abs(wind*0.625)), dircode[wind < 0.0]))


class Water(gameobj.GameObj):
    """Water class defines behavior of the water"""

    def __init__(self, *groups):
        """Init function when creating the object"""
        # Invoke base class init to load all object data
        gameobj.GameObj.__init__(self, 'water', 1.0,
                                 world.scr_scale, *groups)
        # Get the bounding rectangle
        self.rect = self.image['water.png']['surface'].get_rect()
        # Set default animation
        self.SetState('default')
        # Set position
        self.position = 0.0, world.trench_depth - self.rect.height
