#
# Copyright (c) 2007 Patrick Van Oosterwijck
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#

"""Defines the game world"""

import os
import random
import data
import pygame
import gameobj
import objects


# World definitions

trench_width = 800
trench_depth = 6000
view_width = 0
view_height = 0
view_y = 0.0
scr_scale = 1.0
trench = None


# Trench class

class Trench:
    """Defines of the trench from which the balloon has to escape.
    This class holds groups that contain all sprites that make up
    the trench. It also defines trench dimensions and windy regions."""

    def __init__(self):
        """Init function when creating the trench"""
        # Set default scales
        self.obj_scale = 0.75
        self.scr_scale = 1.0
        # Dictionary of block images for the walls
        self.block_image = {}
        # Block sprite group
        self.block_group = gameobj.RenderGameObj()
        # Wind sprite group
        self.wind_group = pygame.sprite.Group()
        # Flag group
        self.flag_group = gameobj.RenderGameObj()

    def Load(self, objscale=None, scrscale=None):
        """Load all images and apply the requested scales"""
        # Save scales if they are provided
        if objscale: self.obj_scale = objscale
        if scrscale: self.scr_scale = scrscale
        # Calculate the current scale, object and screen combined
        scale = self.obj_scale * self.scr_scale
        # Load the block sprites that will make up the walls
        # plus add them flipped and rotated in every direction to
        # make extra block variants
        for filename in os.listdir(data.filepath('blocks')):
            # Open the file and load it
            imagefile = data.load('blocks/' + filename)
            newimage = pygame.image.load(imagefile).convert_alpha()
            # Scale the loaded image
            newimage = pygame.transform.rotozoom(newimage, 0.0, scale)
            # Add all variants to the block image list
            flip = pygame.transform.flip
            rotate = pygame.transform.rotate
            self.block_image[filename] = newimage
            self.block_image[filename+'.lr'] = flip(newimage, True, False)
            self.block_image[filename+'.ud'] = flip(newimage, False, True)
            self.block_image[filename+'.180'] = flip(newimage, True, True)
            self.block_image[filename+'.90'] = rotate(newimage, 90)
            self.block_image[filename+'.270'] = rotate(newimage, 270)

    def Create(self, width, depth, maxoutcrop=None, maxwind=None):
        """Create a trench with the specified dimensions"""
        # Save the trench width and depth
        self.width = width
        self.depth = depth
        # Block overlap range
        overlap_min, overlap_max = 10, 30
        # Set the maximum outcropping
        if maxoutcrop:
            maxoutcrop = width * maxoutcrop
        else:
            maxoutcrop = width * 0.1
        # Shuffle the block images
        block_image_list = self.block_image.keys()
        random.shuffle(block_image_list)
        # Clear all previous blocks
        self.block_group.empty()
        # Create a block list
        blocklist = []
        # Do the left wall, from bottom to top, then the right wall
        # Put all the sprites in the block list
        for side in ['left', 'right']:
            y = float(depth + 100)
            while y > 0:
                newblockimagename = random.choice(block_image_list)
                newblockimage = self.block_image[newblockimagename]
                newrect = newblockimage.get_rect()
                y -= newrect.height - random.uniform(overlap_min, overlap_max)
                if side == 'left':
                    x = -newrect.width + random.uniform(2.0,
                            min(maxoutcrop, 0.9*newrect.width))
                else:
                    x = self.width - random.uniform(2.0,
                            min(maxoutcrop, 0.9*newrect.width))
                newblock = gameobj.StaticObj()
                newblock.image_dict = self.block_image
                newblock.image_name = newblockimagename
                newblock.rect = newrect.move(x, y)
                newblock.collide_rect = newblock.rect
                blocklist.append(newblock)
        # Now shuffle all blocks to randomize Z order
        random.shuffle(blocklist)
        # Add all the blocks to our block group
        self.block_group.add(blocklist)
        # Set default maximum wind if none is provided
        if not maxwind: maxwind = 5.0
        # Clear wind sprite group
        self.wind_group.empty()
        # Clear all previous flags
        self.flag_group.empty()
        # Define wind layers from botton to top
        y = trench_depth
        while y > 0:
            # Create a new layer
            layerheight = random.uniform(150, 400)
            newlayer = gameobj.StaticObj()
            y -= layerheight
            newlayer.rect = pygame.Rect(0, y, trench_width, layerheight)
            newlayer.misc = random.uniform(-maxwind, maxwind)
            self.wind_group.add(newlayer)
            # Create a flag to indicate the wind direction and strength
            newflag = objects.Flag(newlayer.misc)
            newflagrect = newflag.rect
            # Position the new flag on the trench wall
            newflagrect.move_ip(0, y + layerheight/2)
            if newlayer.misc < 0.0:
                newflagrect.move_ip(trench_width-
                        newflagrect.width/self.scr_scale, 0)
            flagblock = None
            for searchblock in self.block_group:
                if newflagrect.colliderect(searchblock.rect):
                    flagblock = searchblock
                    break
            if flagblock:
                newflagrect.top = flagblock.rect.top + \
                        (flagblock.rect.height-newflagrect.height)/2
                if newlayer.misc >= 0.0:
                    newflagrect.left = flagblock.rect.right
                else:
                    newflagrect.left = flagblock.rect.left - newflagrect.width
            newflag.position = newflagrect.topleft
            # Add the new flag to the flag group
            self.flag_group.add(newflag)
            # Create water sprite
            self.water = objects.Water()
